%==========================================================================
% smoothspline.m  version  7.0  (Requires Matlab Curve Fitting Toolbox)
%
% Created:  Brenden Epps, August  9, 2008
% Modified: Brenden Epps, July   28, 2021
%
% [sp,R,E,tol,Rmax,Emax] = smoothspline(x,y,type,mode,plotFlag) returns the 
% best-fit smoothing spline, sp, given a set of presumably-noisy (x,y) data.  
% This spline is "best" in the sense that it has the minimum possible
% error, E, while still having minimal roughness, R.
%
% INPUTS:
%   x       [N,1] vector, independent variable
%   y       [N,1] vector, ..dependent variable
% 
%   type     = 'cubic'   -->  cubic   spline (d^2s/dx^2 == 0 at ends)  (default)
%            = 'quintic' -->  quintic spline (d^3s/dx^3 == 0 at ends) 
% 
%   mode     = 'auto'    -->  automatically select the best curve fit  (default)
%            = 'manual'  -->  manually      select the best curve fit
% 
%   plotFlag = true      -->  show plots in auto mode 
%            = false     -->   no  plots in auto mode (default)
% 
% OUTPUTS:
%   sp    = smoothing spline data structure with standard fields (created by spmak.m):
% 
%               form: 'B-'
%              knots: [1Nk] knot locations,       Nk = Nu+2*No-2
%              coefs: [1Nc] spline coefficients,  Nc = Nu+No-2
%             number: Nc  number of coefficients
%              order: No  spline order (No=2*m == 4 for cubic or 6 for quintic spline)
%                dim: 1   (1-dimensional spline)
%
%         ...and custom fields (appended herein):
% 
%                tol: error tolerance of spline
%                  E: acutal error... of spline
%                  R: roughness...... of spline
%                  x: [1,Nu] x sites with repeats removed
%                  y: [1,Nu] y data  with repeats averaged
%                  s: [1,Nu] spline values,         s = s(x)
%              d1sdx: [1,Nu] spline derivative, d1sdx =   ds/dx   
%              d2sdx: [1,Nu] spline derivative, d2sdx = d^2s/dx^2 
%              d3sdx: [1,Nu] spline derivative, d3sdx = d^3s/dx^3
%
%               where  Nu = number of unique/finite x used to fit spline (possibly smaller than N)
%
%   R(.)     [*,1] roughness... of splines on computed R-E frontier
%   E(.)     [*,1] actual error of splines on computed R-E frontier
%   tol(.)   [*,1] requested error tolerances used to compute R-E frontier  
% 
%   Rmax     upper bound for R, determined from an interpolating spline fit (tol == 0).
%   Emax     upper bound for E, determined from a polynomial fit (for which R == 0).
% 
% ------------------------------
% Theory (adapted from spaps.m):
%
% A smoothing spline s(x) is a piecewise polynomial that approximates
% a given function y(x), perhaps because y contains noise.
% Define spline metrics...
%
%     error:   E(s) := sum_j { w(j)*|y(j) - s(x(j))|^2 }   for j=1,...,length(x)
%
% roughness:   R(s,m) := integral |D^m s(t)|^2 dt    over   min(x) < t < max(x)
%
% where D^m[*] := d^m[*]/dx^m  is the m^th order derivative, where
% m=2 (cubic) and m=3 (quintic).
%
% Herein, the weights w are computed consistent with the Trapezoidal Rule, 
% so the error E is equivalent to evaluating: 
%         integral w(t)*|y(t) - s(t)|^2 dt    over   min(x) < t < max(x).
%
% The Matlab Curve Fitting Toolbox function s = spaps(x,y,tol,type) 
% determines the smoothing spline s that minimizes the roughness R(s,m) 
% while ensuring that the error E(s) is not larger than tol. 
% smoothspline() determines the "best" tol, which allows s to capture  
% the roughness of the true function that the data represents, without
% additional roughness due to the noise in the data.
%
% ------------------------
% Usage notes:
% 
% smoothspline() is intended for use with high-resolution (large N), 
% high-precision (small error) data. In this case, Epps et al. (2010) 
% observed that there exists a critical error tolerance, above which the 
% roughness is relatively constant (due to the "roughness" of the true 
% underlying function), and below which the roughness increases dramatically
% as the spline is forced to pass closer to the noisy data.  
% 
% With mode='manual', smoothspline() computes a series of splines and plots
% an "efficient frontier" of roughness-versus-error of the fitted splines.  
% The user then can choose the spline at the elbow of this frontier.
%
% With mode='auto', smoothspline() automatically determines the tol at the 
% elbow of the roughness-error frontier, using a double-bisection method.
%
%
% smoothspline() is particularly useful for determining the derivatives of
% measured quantities, because derivatives of the fitted spline can be
% computed analytically, with no amplification of measurement noise. For
% examples, see (Truscott et al., 2012) and other papers listed on Google Scholar:
% https://scholar.google.com/scholar?oi=bibs&hl=en&cites=8852866729255769516&as_sdt=5
% 
% 
% smoothspline() requires the MATLAB function "spaps" from the Curve Fitting 
% Toolbox.    https://www.mathworks.com/products/curvefitting.html
% 
% ------------------------
% Note on E verus tol:
%
% Note that the actual error, E, of each spline is slightly different than 
% the error tolerance, tol, prescribed to spaps() and passed to spaps1().  
% The reason is that spaps1() only requires E to be within +/-1% of the 
% desired tol, since iterations in spaps1 terminate when:
%                   if 100*abs(E-tol)<tol, break, end
% 
% Thus, E and tol are close but not necessarily identical.  
%
%-------------------------------------------------------------------------- 
% Example Setup: 
% 
%     % Define a function to analyze
%     x  = 0:0.01:10;   
%     yt = exp(-x).*sin(x);                     % true  y(x)
%     y  = yt +  0.001*randn(size(x));          % noisy y(x)
% 
%     type  = 'cubic';
%     type  = 'quintic';
%
% ---------------------
% Example: Auto mode
% 
%     [sp,R,E,tol] = smoothspline(x,y,type,'auto',true);
% 
% ---------------------
% Example: Manual mode
% 
%     [sp,R,E,tol] = smoothspline(x,y,type,'manual');
%
%   A roughness-error frontier is computed and plotted, and the user is
%   prompted to enter one of the following text commands:
%
%           h ii   hide spline(s)
%           v ii   view spline(s)
%           p      pick new splines to view
%           n (i)  new... error tol range      (default: 20)
%           r (i)  refine error tol resolution (default: 1)
%           d      data plot toggle
%           s i    select spline
%           q      quit
%
% 
%   By way of example,
%
%   >> h125     hides splines 1, 2, and 5.
%
%   >> v12      displays splines 1 and 2.  Vii is the opposite of Hii, and
%               ii can be any one or more of the current spline numbers.
% 
%   >> p        pick new splines: user is prompted to click on points in
%               the Roughness-Error figure, and those splines are shown. 
%  
%   >> n        new error tolerance range:  the user selects two error
%               tolerance values (from the R-E figure), and 20 new splines 
%               in this range are computed.  To specify the number of new
%               splines, say 30, use the command: n30.
%  
%   >> r        refine error tolerance resolution:  the user selects two  
%               error tolerance values (from the R-E figure), and 1 new  
%               spline is inserted between each of the splines in this range.   
%               To specify the number of new splines to insert, say 2, use 
%               the command: r2.
%  
%   >> d        toggles whether the original x,y data is shown.
%
%   >> s3       selects spline 3 for output. This terminates the program.
% 
%   >> q        quit: terminates the program with all outputs given as NaN.
%
% ---------------------
% Example: Postprocessing the output results
%
%     % Check error and roughness of output spline
%     Echeck = trapz(x,       ( y - fnval(sp,x) ).^2 )       % error     of sp
%     Rcheck = trapz(x, fnval(fnder(sp,sp.order/2),x).^2 )   % roughness of sp
%     Echeck - sp.E
%     Rcheck - sp.R
%
%     % Compare output with analytic derivatives and roughness (for the above example):
%     d1ytdx =   -exp(-x).*sin(x) +   exp(-x).*cos(x);    %   dyt/dx
%     d2ytdx = -2*exp(-x).*cos(x);                        % d^2yt/dx^2
%     d3ytdx =  2*exp(-x).*cos(x) + 2*exp(-x).*sin(x);    % d^3yt/dx^3
%     Rt2 = trapz(x, d2ytdx.^2)  % compare to cubic   spline roughness
%     Rt3 = trapz(x, d3ytdx.^2)  % compare to quintic spline roughness 
% 
%     % Plot results
%     figure(5), clf, box on, grid on, loglog(tol,R,'k.','MarkerSize',18)  
%                    hold on,   loglog(sp.tol, sp.R,'r.','MarkerSize',22)
% 
%     figure(4), clf, box on, grid on, plot(x,y,'.b',  x,yt,'k-', sp.x,sp.s,'r--')
%     figure(1), clf, box on, grid on, plot(x,d1ytdx,'k-',  sp.x,sp.d1sdx,'r--')
%     figure(2), clf, box on, grid on, plot(x,d2ytdx,'k-',  sp.x,sp.d2sdx,'r-.')
%     figure(3), clf, box on, grid on, plot(x,d3ytdx,'k-',  sp.x,sp.d3sdx,'r-.')
%
%--------------------------------------------------------------------------
% Reference:
%
%   B.P. Epps, T.T. Truscott, and A.H. Techet, "Evaluating derivatives of
%   experimental data using smoothing splines," Mathematical Methods in 
%   Engineering International Symposium, Coimbra, Portugal, October 2010.
%
%   Truscott, T.T.; Epps, B.P.; and Techet, A.H.; "Unsteady forces on 
%   spheres during free-surface water entry," Journal of Fluid Mechanics,
%   vol. 704, 2012.
%
%   Epps, B.P. (2010) "An impulse framework for hydrodynamic force 
%   analysis: fish propulsion, water entry of spheres, and marine 
%   propellers," Ph.D. thesis, MIT, chapter 5.
% 
%--------------------------------------------------------------------------
% Copyright 2009,2021 Brenden Epps:
% This program is free software.  You can redistribute it and/or modify it
% under the terms of the GNU General Public License version 3, as published
% by the Free Software Foundation.  This program is distributed in the hope
% that it will be useful, but WITHOUT ANY WARRANTY; without even the
% implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
% See the GNU General Public License for more details. You should have 
% received a copy of the GNU General Public License along with this   
% program. If not, see <https://www.gnu.org/licenses/>.
%==========================================================================


function [sp,R,E,tol,Rmax,Emax] = smoothspline(x,y,type,mode,plotFlag)

%--------------------------------------------------------------------------    
% Set default inputs
if nargin < 3,  type     = 'cubic';  end   % cubic smoothing spline
if nargin < 4,  mode     = 'auto';   end   % automatically fit spline
if nargin < 5,  plotFlag = true;     end   % show plots in auto mode

%--------------------------------------------------------------------------    
% Convert string type to numeric value:
switch type
    case {'cubic',  2},  type = 2;  % cubic   smoothing spline
    case {'quintic',3},  type = 3;  % quintic smoothing spline
    otherwise,  error('type must be "cubic" or "quintic".')
end

% Ensure valid mode:
if ~any(strcmp(mode,{'manual','auto','auto2'})),  error('mode must be "auto" or "manual".'), end


%--------------------------------------------------------------------------    
% Perform error checking on the input x and y arrays (and modify them if necessary)
% Compute weights via trapezoidal rule, accounting for repeated x values.
% Modifying x,y here fixes a bug in chckxywp.m that incorrectly handles
% repeated x sites.
%--------------------------------------------------------------------------
[x,y,w] = Check_xy_Compute_w(x,y,type);

%--------------------------------------------------------------------------
% Find upper bounds for error and roughness, used to initialize stencil
%--------------------------------------------------------------------------
% Find upper bound for error tolerance, tol, for which roughness is zero.
%   type == 3  ---->  type-1 == 2 is a quadratic fit
%   type == 2  ---->  type-1 == 1 is a linear    fit
Emax = trapz(x, ( y - polyval(polyfit(x,y,type-1),x) ).^2 );

% Round down so that roughness of the fitted spline is nonzero (prevents bugs)
Emax_mod = 10^(floor(10*log10(Emax))/10);

% Find upper bound for R using interpolating spline (tol == 0)
Rmax = FitSpline_ComputeRoughness(x,y,type, 0);


    
%--------------------------------------------------------------------------
switch mode
case 'auto'
    %======================================================================
    % AUTO mode
    %======================================================================
    C = 5;  % number of splines in stencil (must be ODD)
    
    %----------------------------------------------------------------------
    % Define initial double-bisection stencil...via R-T frontier 
    %----------------------------------------------------------------------
    disp('Initializing double-bisection stencil...')
    
    % Define range of error tolerance values
    T = 10.^[log10(Emax_mod) + [-12 : 0.5 : 0]]';
    
    Ne = length(T);  % number of splines to evaluate
    
    E = zeros(Ne,1);  % error
    R = zeros(Ne,1);  % roughness

    for i = Ne : -1 : 1
        fprintf('  ...Evaluating spline %2d of %2d. \n',1+Ne-i,Ne),

        [R(i),~,E(i)] = FitSpline_ComputeRoughness(x,y,type,T(i));
        
        % Stop computing splines if R(i) is nearly Rmax
        if i < Ne+1-C && abs(Rmax - R(i)) < 0.1*Rmax
            T = T(i:Ne);
            E = E(i:Ne);
            R = R(i:Ne);
            Ne = length(T);
            fprintf('  ...Roughness nearly maximum value.  Continuing...\n')
            break,
        end
    end
    
    % Ensure R is nonzero (should at most change R(Ne) for which tol=Emax_mod):
    R = max(R, 1e-100);
    
    %----------------------------------------------------------------------
    % Record initial frontier in "all values" arrays
    %----------------------------------------------------------------------
    Tall = zeros(200+Ne,1);   Tall(1:Ne) = T(:);
    Eall = zeros(200+Ne,1);   Eall(1:Ne) = E(:);
    Rall = zeros(200+Ne,1);   Rall(1:Ne) = R(:);
    Nall = Ne;  % current number of entries in "all values" arrays    

    %----------------------------------------------------------------------
    % Initialize plots
    %----------------------------------------------------------------------
    if plotFlag
        DefaultAxes_FontSize = 18;
        DefaultLabelFontSize = 18;
        psize = 26;  % period size
        Green = [0  0.7  0];

        %------------------------------------------------------------------
        % Plot (T,R) - fixed view
        Hfv = figure(14);
            set(Hfv,'Position',[1  1   560   420])
            clf, hold on, box on, grid on,
            HHmax = plot([T(1),T(end)],[Rmax,Rmax],'k--');
 
            HHfv_a = plot(NaN,NaN,'k.','MarkerSize',psize);
            HHfv_5 = plot(NaN,NaN,'r.-','MarkerSize',psize,'LineWidth',1);
            HHfv_1 = plot(NaN,NaN, '.','MarkerSize',psize,'color',Green);

            xlabel('Error tolerance, $tol$','Fontsize',DefaultLabelFontSize,'Interpreter','latex')
            ylabel('Roughness, $R$'        ,'Fontsize',DefaultLabelFontSize,'Interpreter','latex')
            set(gca,'FontSize',DefaultAxes_FontSize,'TickLabelInterpreter','latex')
            set(gca,'XScale','log','YScale','log')

            axis([T(1) T(end) R(end-1)/5 Rmax]) 
            
        %------------------------------------------------------------------
        % Plot (T,R) - variable zoom 
        Hvz = figure(15);
            set(Hvz,'Position',[561  1   560   420])
            clf, hold on, box on, grid on,

            HHvz_5 = plot(NaN,NaN,'r.-','MarkerSize',psize,'LineWidth',1);
            HHvz_1 = plot(NaN,NaN, '.','MarkerSize',psize,'color',Green);

            xlabel('Error tolerance, $tol$','Fontsize',DefaultLabelFontSize,'Interpreter','latex')
            ylabel('Roughness, $R$'        ,'Fontsize',DefaultLabelFontSize,'Interpreter','latex')
            set(gca,'FontSize',DefaultAxes_FontSize,'TickLabelInterpreter','latex')
            set(gca,'XScale','log','YScale','log')          
    end
    

    %----------------------------------------------------------------------
    % Find (T,R) point of max positive curvature
    %----------------------------------------------------------------------
    Rpp = Compute_R_E_Curvature(R,T);

    [~,Ibest] = max(Rpp);  % index of point with max positive curvature

    Chalf = (C-1)/2;  % "half" of stencil size
    
    if  Ibest > Ne - Chalf
        Ibest = Ne - Chalf;  
    end
    
    %----------------------------------------------------------------------
    % Choose initial stencil surrounding best point
    %----------------------------------------------------------------------
    T = T(Ibest-Chalf:Ibest+Chalf); % T of each spline fit choice
    E = E(Ibest-Chalf:Ibest+Chalf); % E of each spline fit choice
    R = R(Ibest-Chalf:Ibest+Chalf); % R of each spline fit choice    
    
    %----------------------------------------------------------------------
    % Perform double-bisection iterations until convergence
    %----------------------------------------------------------------------
    done  = false;
    count = 0;

    while (~done) &&  (count < 100)
        count = count + 1;
        
        % Find curvature of interior stencil points
        Rpp = Compute_R_E_Curvature(R,T);
        
        % Find stencil point with max curvature
        [~,p] = max(Rpp);

        %------------------------------------------------------------------
        % Update plot data
        %------------------------------------------------------------------
        if plotFlag
            %--------------------------------------------------------------
            % Plot (T,R) - fixed view
            HHfv_a.XData = Tall(1:Nall);
            HHfv_a.YData = Rall(1:Nall);

            HHfv_5.XData = T;
            HHfv_5.YData = R;

            HHfv_1.XData = T(p);
            HHfv_1.YData = R(p);

            %--------------------------------------------------------------
            % Plot (T,R) - variable zoom 
            HHvz_5.XData = T;
            HHvz_5.YData = R;

            HHvz_1.XData = T(p);
            HHvz_1.YData = R(p);            

            drawnow,
            pause(0.1),
        end                        

        %------------------------------------------------------------------
        % Update stencil and evaluate roughness of two new splines
        %------------------------------------------------------------------
        Told = T;
        T(1) =      Told(p-1);
        T(2) = sqrt(Told(p-1)*Told(p));
        T(3) =                Told(p);
        T(4) = sqrt(Told(p+1)*Told(p));
        T(5) =      Told(p+1);

        Eold = E;
        E(1) = Eold(p-1); 
        E(3) = Eold(p);
        E(5) = Eold(p+1);

        Rold = R;
        R(1) = Rold(p-1); 
        R(3) = Rold(p);
        R(5) = Rold(p+1);
        
       [R(2),~,E(2)] = FitSpline_ComputeRoughness(x,y,type,T(2)); 
       [R(4),~,E(4)] = FitSpline_ComputeRoughness(x,y,type,T(4));

        %------------------------------------------------------------------
        % Record new values in "all values" arrays
        Tall(Nall+1) = T(2);   Rall(Nall+1) = R(2);   Eall(Nall+1) = E(2);
        Tall(Nall+2) = T(4);   Rall(Nall+2) = R(4);   Eall(Nall+2) = E(4);
        Nall = Nall + 2;
        
        %------------------------------------------------------------------
        % Check for convergence (if 0.1% difference between stencil points)
        conv = abs((log(T(4)) - log(T(3)))/log(T(4)));
        
        fprintf('Iteration %2d:  R = %9.4e,  tol = %9.4e,  dtol/tol = %6.4f ',count, R(4),T(4), conv)
        
        if conv < 0.001 
            done = true;
            fprintf(' (converged)')
        end
        fprintf('\n')
        
    end  % double-bisection while loop  
  
    %----------------------------------------------------------------------
    % Create selected smoothing spline... using error level T(4)
    sp     = spaps(x,y,T(4),type);
    sp.tol = T(4);
    sp.E   = ComputeError(x,y,sp);
    sp.R   = ComputeRoughness(x,type,sp);

     
%======================================================= END AUTO mode code        
case 'auto2'
    %======================================================================
    % AUTO2 mode
    %======================================================================
    %----------------------------------------------------------------------
    % Define initial double-bisection stencil, T(1:5)
    %----------------------------------------------------------------------
    disp('Initializing double-bisection stencil...')
            
    % Estimate critical error, Ecr
    %   Ecr   ~ N e^2 dt
    %   R2max = 36 N e^2 dt^-3 ~ 36 Ecr dx^-4   (cubic   splines)
    %   R3max = 31 N e^2 dt^-5 ~ 31 Ecr dx^-6   (quintic splines)
    dx  = mean( diff(x) );

    if     type == 2,  Ecr = Rmax / (36 * dx^-4);   % (cubic   splines) 
    elseif type == 3,  Ecr = Rmax / (31 * dx^-6);   % (quintic splines)  
    end
    
    % Set initial stencil
    T  = zeros(5,1); % error tolerance
    E  = zeros(5,1); % actual error
    R  = zeros(5,1); % roughness

    T(1) = 1e-14;            % note: machine zero is 2.2204e-16
    T(2) = sqrt(Ecr*1e-14);  % log(T2) = 1/2 (log(Ecr)+log(1e-14))
    T(3) =      Ecr;
    T(4) = sqrt(Ecr*Emax_mod);
    T(5) = Emax_mod;

    %----------------------------------------------------------------------
    % Compute roughness of initial stencil, and augment stencil if necessary
    %----------------------------------------------------------------------
    % Compute roughness of point T(1)
    [R(1),~,E(1)] = FitSpline_ComputeRoughness(x,y,type, T(1));

    if R(1) == 0
        error('Roughness is zero with tol(1) = 1e-14.')
    end

    % Compute roughness at T(3), and adjust T(3) if necessary to achieve non-zero roughness
    Rflag = 0;

    while Rflag == 0

        [R(3),~,E(3)] = FitSpline_ComputeRoughness(x,y,type, T(3));

        if R(3) == 0
            T(3) = sqrt(T(3)*1e-14);  % make log(T(3)) half as much
        else
            Rflag = 1;
        end
    end

    % Compute roughness of point (2)
     T(2) = sqrt(T(3)*1e-14);  % log(T(2)) = 1/2 (log(T(3))+log(1e-14))
    [R(2),~,E(2)] = FitSpline_ComputeRoughness(x,y,type, T(2));


    % Compute roughness of point (5), and ensure R(5) is not zero
    [R(5),~,E(5)] = FitSpline_ComputeRoughness(x,y,type, T(5));

    R(5) = max( R(5), 10^-100 );

    
    % Compute roughness at T(4), and adjust T(4) if necessary to achieve non-zero roughness
    T(4) = sqrt(T(3)*T(5));  % log(T(4)) = 1/2 (log(T(3))+log(T(5)))

    Rflag = 0;

    while Rflag == 0
        [R(4),~,E(4)] = FitSpline_ComputeRoughness(x,y,type, T(4));

        if R(4) == 0
            T(4) = sqrt(T(3)*T(4));
        else
            Rflag = 1;
        end
    end

    %----------------------------------------------------------------------
    % Ensure that the initial stencil has a point of positive curvature
    %----------------------------------------------------------------------
    Rpp = Compute_R_E_Curvature(R,T);

    % If no point with positive curvature, adjust the T(3) and try again
    if ~any( Rpp > 0 )
         T(3) = sqrt(T(3)*T(4));
        [R(3),~,E(3)] = FitSpline_ComputeRoughness(x,y,type, T(3));
    
        Rpp = Compute_R_E_Curvature(R,T);
    end
    
    % If no point with positive curvature, adjust the T(2) and T(4) and try again
    if ~any( Rpp > 0 )
        
        T(2) = sqrt(T(3)*T(2));
        T(4) = sqrt(T(3)*T(4));
        
        [R(2),~,E(2)] = FitSpline_ComputeRoughness(x,y,type, T(2));
        [R(4),~,E(4)] = FitSpline_ComputeRoughness(x,y,type, T(4));
        
        Rpp = Compute_R_E_Curvature(R,T);
    end
    
    % If no point with positive curvature, then suggest manual mode
    if ~any( Rpp > 0 )
        error('Double-bisection method stalled. Try another mode.')
    end 
    
    %----------------------------------------------------------------------
    % Record initial stencil in "all values" arrays
    %----------------------------------------------------------------------
    Tall = zeros(205,1);   Tall(1:5) = T(1:5);
    Eall = zeros(205,1);   Eall(1:5) = E(1:5);
    Rall = zeros(205,1);   Rall(1:5) = R(1:5);
    Nall = 5;  % current number of entries in "all values" arrays
    
    %----------------------------------------------------------------------
    % Initialize plots
    %----------------------------------------------------------------------
    if plotFlag
        DefaultAxes_FontSize = 18;
        DefaultLabelFontSize = 18;
        psize = 26;  % period size
        Green = [0  0.7  0];

        %------------------------------------------------------------------
        % Plot (T,R) - fixed view
        Hfv = figure(14);
            set(Hfv,'Position',[1  1   560   420])
            clf, hold on, box on, grid on,
            HHmax = plot([T(1),T(end)],[Rmax,Rmax],'k--');
 
            HHfv_a = plot(NaN,NaN,'k.','MarkerSize',psize);
            HHfv_5 = plot(NaN,NaN,'r.-','MarkerSize',psize,'LineWidth',1);
            HHfv_1 = plot(NaN,NaN, '.','MarkerSize',psize,'color',Green);

            xlabel('Error tolerance, $tol$','Fontsize',DefaultLabelFontSize,'Interpreter','latex')
            ylabel('Roughness, $R$'        ,'Fontsize',DefaultLabelFontSize,'Interpreter','latex')
            set(gca,'FontSize',DefaultAxes_FontSize,'TickLabelInterpreter','latex')
            set(gca,'XScale','log','YScale','log')

            axis([T(1) T(end) R(end-1) Rmax]) 
            
        %------------------------------------------------------------------
        % Plot (T,R) - variable zoom 
        Hvz = figure(15);
            set(Hvz,'Position',[561  1   560   420])
            clf, hold on, box on, grid on,

            HHvz_5 = plot(NaN,NaN,'r.-','MarkerSize',psize,'LineWidth',1);
            HHvz_1 = plot(NaN,NaN, '.','MarkerSize',psize,'color',Green);

            xlabel('Error tolerance, $tol$','Fontsize',DefaultLabelFontSize,'Interpreter','latex')
            ylabel('Roughness, $R$'        ,'Fontsize',DefaultLabelFontSize,'Interpreter','latex')
            set(gca,'FontSize',DefaultAxes_FontSize,'TickLabelInterpreter','latex')
            set(gca,'XScale','log','YScale','log')          
    end
    
    %----------------------------------------------------------------------
    % Perform double-bisection iterations until convergence
    %----------------------------------------------------------------------
    done  = false;
    count = 0;
    
    while (~done) &&  (count < 100)
        count = count + 1;
        
        % Find curvature of interior stencil points
        Rpp = Compute_R_E_Curvature(R,T);
        
        % Find stencil point with max curvature
        [~,p] = max(Rpp);

        %------------------------------------------------------------------
        % Update plot data
        %------------------------------------------------------------------
        if plotFlag
            %--------------------------------------------------------------
            % Plot (T,R) - fixed view
            HHfv_a.XData = Tall(1:Nall);
            HHfv_a.YData = Rall(1:Nall);

            HHfv_5.XData = T;
            HHfv_5.YData = R;

            HHfv_1.XData = T(p);
            HHfv_1.YData = R(p);

            %--------------------------------------------------------------
            % Plot (T,R) - variable zoom 
            HHvz_5.XData = T;
            HHvz_5.YData = R;

            HHvz_1.XData = T(p);
            HHvz_1.YData = R(p);            

            drawnow,
            pause(0.1),
          % pause,
        end                        

        %------------------------------------------------------------------
        % Update stencil and evaluate roughness of two new splines
        %------------------------------------------------------------------
        Told = T;
        T(1) =      Told(p-1);
        T(2) = sqrt(Told(p-1)*Told(p));
        T(3) =                Told(p);
        T(4) = sqrt(Told(p+1)*Told(p));
        T(5) =      Told(p+1);

        Eold = E;
        E(1) = Eold(p-1); 
        E(3) = Eold(p);
        E(5) = Eold(p+1);

        Rold = R;
        R(1) = Rold(p-1); 
        R(3) = Rold(p);
        R(5) = Rold(p+1);
        
       [R(2),~,E(2)] = FitSpline_ComputeRoughness(x,y,type,T(2)); 
       [R(4),~,E(4)] = FitSpline_ComputeRoughness(x,y,type,T(4));
        

        % Record new values in "all values" arrays
        Tall(Nall+1) = T(2);   Rall(Nall+1) = R(2);   Eall(Nall+1) = E(2);
        Tall(Nall+2) = T(4);   Rall(Nall+2) = R(4);   Eall(Nall+2) = E(4);
        Nall = Nall + 2;
        
        % Check for convergence (if 0.1% difference between stencil points)
        conv = abs((log(T(4)) - log(T(3)))/log(T(4)));
        
        fprintf('Iteration %2d:  R = %9.4e,  tol = %9.4e,  dtol/tol = %6.4f ',count, R(4),T(4), conv)
        
        if conv < 0.001 
            done = true;
            fprintf(' (converged)')
        end
        fprintf('\n')
        
    end  % double-bisection while loop  
  
    %----------------------------------------------------------------------
    % Create selected smoothing spline... using error level T(4)
    sp     = spaps(x,y,T(4),type);
    sp.tol = T(4);
    sp.E   = ComputeError(x,y,sp);
    sp.R   = ComputeRoughness(x,type,sp);
     
%====================================================== END AUTO2 mode code     
case 'manual'  
    %======================================================================
    % MANUAL mode
    %======================================================================  
    N = length(x);  % number of data sites
    C = 5;          % number of choices of spline fit

    %----------------------------------------------------------------------
    % Generate R-T frontier for smoothing splines 
    %----------------------------------------------------------------------
    fprintf('\n'),
    fprintf('Generating (Error,Roughness) frontier for smoothing splines ... \n'),

    % Define range of error tolerance values
    T = 10.^[log10(Emax_mod) + [-12 : 0.1 : 0]]';

    Ne = length(T);  % number of splines to evaluate
    
    E = zeros(Ne,1);  % error
    R = zeros(Ne,1);  % roughness

    % define struct for B-spline data
    S       = struct('form','B-','knots',[],'coefs',[],'number',[],'order',[],'dim',1);
    S(Ne,1) = S(1);  % allocate memory for array of structs
    
    for i = Ne : -1 : 1
        fprintf('Evaluating spline %2d of %2d. \n',1+Ne-i,Ne),

        [R(i),S(i),E(i)] = FitSpline_ComputeRoughness(x,y,type,T(i));
        
        % Stop computing splines if R(i) is nearly Rmax
        if i < Ne+1-C && abs(Rmax - R(i)) < 0.1*Rmax
            T = T(i:Ne);
            E = E(i:Ne);
            R = R(i:Ne);
            S = S(i:Ne);
            Ne = length(T);
            fprintf('Roughness nearly maximum value.  Continuing...\n')
            break,
        end
    end
    
    % Ensure R is nonzero (should at most change R(Ne) for which tol=Emax_mod):
    R = max(R, 1e-100);  
    
    %----------------------------------------------------------------------
    % Find (T,R) point of max positive curvature
    %----------------------------------------------------------------------
    Rpp = Compute_R_E_Curvature(R,T);

    [~,Ibest] = max(Rpp);  % index of point with max positive curvature
  
    if  Ibest > Ne + 1 - C
        Ibest = Ne + 1 - C;  
    end
    
    %----------------------------------------------------------------------
    % Generate C choices of possible spline fits
    %----------------------------------------------------------------------
    fprintf('\n'),
    fprintf('Generating possible smoothing spline choices...\n'),
     
    DC   = true(1,C);             % Display Curves    flag
    DD   = true;                     % Display Data      flag

    Tsfc = T(Ibest:Ibest+C-1); % T of each spline fit choice
    Esfc = E(Ibest:Ibest+C-1); % E of each spline fit choice
    Rsfc = R(Ibest:Ibest+C-1); % R of each spline fit choice
    Ssfc = S(Ibest:Ibest+C-1); % S of each spline fit choice
    
    yy   = zeros(C,N);         % y(x)
    y1   = zeros(C,N);         %   dy/dx
    y2   = zeros(C,N);         % d^2y/dx^2
    y3   = zeros(C,N);         % d^3y/dx^3

    for i = 1:C
        yy(i,:) = fnval(      Ssfc(i)   ,x);
        y1(i,:) = fnval(fnder(Ssfc(i),1),x);
        y2(i,:) = fnval(fnder(Ssfc(i),2),x);
        y3(i,:) = fnval(fnder(Ssfc(i),3),x);
    end

    %----------------------------------------------------------------------
    % Record initial frontier in "all values" arrays
    %----------------------------------------------------------------------
    Tall = zeros(200+Ne,1);   Tall(1:Ne) = T(:);
    Eall = zeros(200+Ne,1);   Eall(1:Ne) = E(:);
    Rall = zeros(200+Ne,1);   Rall(1:Ne) = R(:);
    Nall = Ne;  % current number of entries in "all values" arrays   
    
    %----------------------------------------------------------------------
    %% Plot results
    %----------------------------------------------------------------------
    disp(' '),
    disp('Plotting results...'),
    
    DefaultAxes_FontSize = 18;
    DefaultLabelFontSize = 18;
    psize = 26;  % period size
    osize =  6;
    
    LegendStyle = {'Fontsize',DefaultLabelFontSize,'Interpreter','latex','Location','best'};
    
    % Allocate memory for plot handles
    HHyy = zeros(1,C);
    HHy1 = zeros(1,C);
    HHy2 = zeros(1,C);
    HHy3 = zeros(1,C);
    HHro = zeros(1,C);

    % Define distinguisable X11 colors:
    Red         = [1.0000         0         0];
    Green       = [0         0.7000         0];
    DodgerBlue  = [0.1172    0.5625    1.0000];
    BlueViolet  = [0.5391    0.1680    0.8828];
    DeepPink    = [1.0000    0.0781    0.5742];
    color_orange= [1.0000    0.5000    0.0000]; 

    color_keynote_blue2   = [  3 101 192]/255;
    color_keynote_green2  = [  0 136  43]/255;
    color_keynote_orange2 = [222 106  16]/255;

    % Define color matrix      
    CLR = [ Red;          % 1 
            DodgerBlue;   % 2
            color_orange; % 3
            BlueViolet;   % 4
            Green;        % 5
            DeepPink;     % 6
            color_keynote_blue2;    % 7
            color_keynote_orange2;  % 8
            color_keynote_green2];  % 9        

    CLRlegend_all = {'(1)','(2)','(3)','(4)','(5)','(6)','(7)','(8)','(9)'};           

    CLRlegend = CLRlegend_all(1:C);
    DATlabel = {'data'};  

    %----------------------------------------------------------------------
    % Plot (T,R)
    Hro = figure(5);
        set(Hro,'Position',[561  495   560   420])
        clf, hold on, box on, grid on,
        HHmax = plot([T(1),T(end)],[Rmax,Rmax],'k--');
        
        HHer = plot(T,R,'k.','MarkerSize',psize);
        for i = 1:C
            HHro(i) = plot(Tsfc(i),Rsfc(i),'.','color',CLR(i,:),'MarkerSize',psize+2);
        end
  
        xlabel('Error tolerance, $tol$','Fontsize',DefaultLabelFontSize,'Interpreter','latex')
        ylabel('Roughness, $R$'        ,'Fontsize',DefaultLabelFontSize,'Interpreter','latex')
        set(gca,'FontSize',DefaultAxes_FontSize,'TickLabelInterpreter','latex')
        set(gca,'XScale','log','YScale','log')
        
        set(gca,'XLim',[T(1) T(end)])
        legend(HHro(DC),CLRlegend(DC),LegendStyle{:})

    %----------------------------------------------------------------------
    % Plot (X,Y)
    Hy = figure(4);
        set(Hy,'Position',[1     495   560   420]) 
        clf, hold on, box on, grid on,
        plot(x,0*x,'k-','linewidth',1)
        HHdata = plot(x,y,'ko','MarkerSize',osize);
        for i = 1:C
            HHyy(i) = plot(x,yy(i,:),'color',CLR(i,:),'linewidth',1);    
        end
        set(gca,'XLim',[x(1) x(end)])
        set(gca,'FontSize',DefaultAxes_FontSize,'TickLabelInterpreter','latex')
        xlabel('$x$',                 'Fontsize',DefaultLabelFontSize,'Interpreter','latex')
        ylabel('$y{\scriptstyle(x)}$','Fontsize',DefaultLabelFontSize,'Interpreter','latex')
        legend([HHdata(DD),HHyy(DC)],[DATlabel(DD),CLRlegend(DC)],LegendStyle{:})
        
    %----------------------------------------------------------------------
    % Plot (X,Yp)
    Hy1 = figure(1);
        set(Hy1,'Position',[1     1   560   420])
        clf, hold on, box on, grid on,
        plot(x,0*x,'k-','linewidth',1)
        for i = 1:C    
            HHy1(i) = plot(x,y1(i,:),'color',CLR(i,:),'linewidth',1);
        end
        set(gca,'XLim',[x(1) x(end)])
        set(gca,'FontSize',DefaultAxes_FontSize,'TickLabelInterpreter','latex')
        xlabel('$x$',    'Fontsize',DefaultLabelFontSize,'Interpreter','latex')
        ylabel('$dy/dx$','Fontsize',DefaultLabelFontSize,'Interpreter','latex')
        legend(HHy1(DC),CLRlegend(DC),LegendStyle{:})
    
    %----------------------------------------------------------------------
    % Plot (X,Ypp)
    Hy2 = figure(2);
        set(Hy2,'Position',[561     1   560   420])
        clf, hold on, box on, grid on,
        plot(x,0*x,'k-','linewidth',1)
        for i = 1:C
            HHy2(i) = plot(x,y2(i,:),'color',CLR(i,:),'linewidth',1);
        end
        set(gca,'XLim',[x(1) x(end)])
        set(gca,'FontSize',DefaultAxes_FontSize,'TickLabelInterpreter','latex')
        xlabel('$x$',        'Fontsize',DefaultLabelFontSize,'Interpreter','latex')
        ylabel('$d^2y/dx^2$','Fontsize',DefaultLabelFontSize,'Interpreter','latex')
        legend(HHy2(DC),CLRlegend(DC),LegendStyle{:})
    
    %----------------------------------------------------------------------
    % Plot (X,Yppp)
    Hy3 = figure(3);
        set(Hy3,'Position',[1121           1         560         420])
        clf, hold on, box on, grid on,
        plot(x,0*x,'k-','linewidth',1)
        for i = 1:C
            HHy3(i) = plot(x,y3(i,:),'color',CLR(i,:),'linewidth',1);
        end
        set(gca,'XLim',[x(1) x(end)])
        set(gca,'FontSize',DefaultAxes_FontSize,'TickLabelInterpreter','latex')
        xlabel('$x$',        'Fontsize',DefaultLabelFontSize,'Interpreter','latex')
        ylabel('$d^3y/dx^3$','Fontsize',DefaultLabelFontSize,'Interpreter','latex')
        legend(HHy3(DC),CLRlegend(DC),LegendStyle{:})
    
    %----------------------------------------------------------------------
    % Bring T-R figure to focus
    figure(Hro),
        

    %----------------------------------------------------------------------
    %% Manually select best smoothing spline
    %----------------------------------------------------------------------
    done = false;

    while (~done)
        
        disp(' ')
        disp('Input an action: ')
        disp('  h ii   hide spline(s)')
        disp('  v ii   view spline(s)') 
        disp('  p      pick new splines to view')
        disp('  n (i)  new... error tol range      (default: 20)')
        disp('  r (i)  refine error tol resolution (default: 1)')
        disp('  d      data plot toggle')
        disp('  s i    select spline')
        disp('  q      quit')
        disp(' ')

        pause(0.0001),
        action = input('  >> ','s');
        
        if isempty(action)    
            disp('No action given...try again.')
            continue, 
        end

        switch action(1)            
        %------------------------------------------------------------------
        case {'s','S'}
            if length(action) == 1
                Cbest = input(sprintf('Select curve (1-%d): ',C));
            else
                Cbest = str2num(action(2));
            end    

            DC        = false(1,C);
            DC(Cbest) = true;

            done = true;

            % Update which curve fits are displayed:
            UpdatePlotVisibility(Hy,HHyy,HHdata, Hy1,HHy1, Hy2,HHy2, Hy3,HHy3, Hro, DC,DD,DATlabel,CLRlegend,LegendStyle);
 
        %------------------------------------------------------------------
        case {'h','H'}
            if length(action) == 1
                RMcurve = input('Hide which curve(s)? ','s');
            else
                RMcurve = action(2:end);
            end

            for i = 1:length(RMcurve)
                DC(str2num(RMcurve(i))) = false;
            end

            % Update which curve fits are displayed:
            UpdatePlotVisibility(Hy,HHyy,HHdata, Hy1,HHy1, Hy2,HHy2, Hy3,HHy3, Hro, DC,DD,DATlabel,CLRlegend,LegendStyle);

        %------------------------------------------------------------------
        case {'v','V'}
            if length(action) == 1
                ADDcurve = input('View which curve(s)? ','s');
            else
                ADDcurve = action(2:end);
            end    

            for i = 1:length(ADDcurve)
                DC(str2num(ADDcurve(i))) = true;
            end    

            % Update which curve fits are displayed:
            UpdatePlotVisibility(Hy,HHyy,HHdata, Hy1,HHy1, Hy2,HHy2, Hy3,HHy3, Hro, DC,DD,DATlabel,CLRlegend,LegendStyle);

        %------------------------------------------------------------------
        case {'p','P'}
            disp(' ')
            disp('Click on points to view curve fits.')
            disp('Press "return" when done.')

            %--------------------------------------------------------------
            % Select new (T,R) point(s)
            figure(Hro);
            
            [Tdes,Rdes] = ginput;

            logT = log(T);
            logR = log(R);
            
            Ides = zeros(length(Tdes),1);
            
            % Find the nearest point
            for i = 1:length(Ides)
                [~, Ides(i)] = min( (logT - log(Tdes(i))).^2 + (logR - log(Rdes(i))).^2 );
            end

            % Ensure indices are unique (and sorted in ascending order)
            Ides = unique(Ides);


            %--------------------------------------------------------------
            % Delete existing plot data
            delete([HHyy,HHy1,HHy2,HHy3,HHro])

            %--------------------------------------------------------------
            % If C has changed, then need to reallocate memory
            if C ~= length(Ides)
                C = length(Ides); 

                if C > 9
                    disp(' ')
                    disp('Note: max display of 9 curves.')
                    C    = 9;
                    Ides = Ides(1:C);
                end

                HHyy    = zeros(1,C);
                HHy1    = zeros(1,C);
                HHy2    = zeros(1,C);
                HHy3    = zeros(1,C);
                HHro    = zeros(1,C);
                yy      = zeros(C,N);  % y(x)
                y1      = zeros(C,N);  % dy/dx
                y2      = zeros(C,N);  % d^2y/dx^2
                y3      = zeros(C,N);  % d^3y/dx^3
                Tsfc    = zeros(C,1);
                Esfc    = zeros(C,1);
                Rsfc    = zeros(C,1);  
                Ssfc    = Ssfc(1);
                Ssfc(C) = Ssfc(1);
            end
            
            %--------------------------------------------------------------
            % Compute curves for new spline fit choices
            for i = 1:C
                Tsfc(i) = T( Ides(i) );
                Esfc(i) = E( Ides(i) );
                Rsfc(i) = R( Ides(i) );
                Ssfc(i) = S( Ides(i) );
                
                yy(i,:) = fnval(      Ssfc(i)   ,x);
                y1(i,:) = fnval(fnder(Ssfc(i),1),x);
                y2(i,:) = fnval(fnder(Ssfc(i),2),x);
                y3(i,:) = fnval(fnder(Ssfc(i),3),x);
            end
            
            %--------------------------------------------------------------
            % Plot all new curves
            CLRlegend = CLRlegend_all(1:C); % Legend
            DC        = true(1,C);          % Display Curves flag

            figure(Hy)
                for i = 1:C
                    HHyy(i) = plot(x,yy(i,:),'color',CLR(i,:),'linewidth',1);    
                end
                legend([HHdata(DD),HHyy(DC)],[DATlabel(DD),CLRlegend(DC)],LegendStyle{:})
                
            figure(Hy1)
                for i = 1:C    
                    HHy1(i) = plot(x,y1(i,:),'color',CLR(i,:),'linewidth',1);
                end
            	legend(HHy1(DC),CLRlegend(DC),LegendStyle{:})
                
            figure(Hy2)
                for i = 1:C
                    HHy2(i) = plot(x,y2(i,:),'color',CLR(i,:),'linewidth',1);
                end
            	legend(HHy2(DC),CLRlegend(DC),LegendStyle{:})
                
            figure(Hy3)    
                for i = 1:C
                    HHy3(i) = plot(x,y3(i,:),'color',CLR(i,:),'linewidth',1);
                end
            	legend(HHy3(DC),CLRlegend(DC),LegendStyle{:})

            figure(Hro)
                for i = 1:C
                    HHro(i) = plot(Tsfc(i),Rsfc(i),'.','color',CLR(i,:),'MarkerSize',psize+2);
                end
                legend(HHro(DC),CLRlegend(DC),LegendStyle{:})
            %--------------------------------------------------------------
        
        %------------------------------------------------------------------
        case {'r','R'}
            if length(action) == 1
                Mi = 1;  % number of new points to insert
            else
                Mi = str2num(action(2:end));
            end
            
            %--------------------------------------------------------------
            % Select new T bounds
            disp(' ')
            disp('Mark lower and upper bounds for tol refinement.')
            
            figure(Hro);
            
            [Tdes,Rdes] = ginput(2);

            % Ensure T are unique (and sorted in ascending order)
            Tdes = unique(Tdes);
            
            if length(Tdes)==1,  error('Must choose different tol values.'), end

            %--------------------------------------------------------------
            % Find T values in range:  Tdes(1) < T < Tdes(2)
            Ikeep = find( Tdes(1) < T  &  T < Tdes(2) );
            
            No = length(Ikeep);  % number of old splines to keep
            Na = Mi*(No-1);      % number of new splines to add
            Nnew = No + Na;        % number of new splines in total

           
            %--------------------------------------------------------------
            % Generate refined R-T frontier for smoothing splines 
            %--------------------------------------------------------------
            fprintf('\n'),
            fprintf('Refining roughness-error frontier... \n'),

            % Allocate memory
            Tnew = zeros(Nnew,1);  % error
            Enew = zeros(Nnew,1);  % error
            Rnew = zeros(Nnew,1);  % roughness

            % define struct for B-spline data
            Snew         = struct('form','B-','knots',[],'coefs',[],'number',[],'order',[],'dim',1);
            Snew(Nnew,1) = Snew(1);  % allocate memory for array of structs

            in = 0; % running index in new arrays
            ca = 0; % running count of new splines
            
            % Evaluate each spline
            for ik = 1:No  % for each segment in which to add splines
                in = in + 1;
                io = Ikeep(ik);  % index in old arrays
                
                % fprintf('in = %2d,  io = %2d \n',in,io)
                
                Tnew(in) = T(io);
                Enew(in) = E(io);
                Rnew(in) = R(io);
                Snew(in) = S(io);

                if ik < No
                    for ii = 1:Mi  % for each spline to insert
                        in = in + 1;
                        ca = ca + 1;
                        frac = ii / (Mi+1);  % fraction of distance across segment
                        
                        % fprintf('in = %2d,  io = %d,  ii = %d, frac = %6.4f \n',in,io,ii,frac)
                        
                        fprintf('Evaluating spline %2d of %2d. \n',ca,Na),

                        % linearly interpolate tolerance
                        Tnew(in) = 10^( (1-frac)*log10(T(io)) + frac*log10(T(io+1)) );  
                        
                        % evaluate spline
                        [Rnew(in),Snew(in),Enew(in)] = FitSpline_ComputeRoughness(x,y,type,Tnew(in));
                    end
                end
            end
            
            % Insert new values into old arrays
            T = [T; Tnew];
            E = [E; Enew];
            R = [R; Rnew];
            S = [S; Snew];
            
            % Only keep unique splines, and sort arrays:
            [~,ind] = unique(T);   
            T = T(ind);
            E = E(ind);
            R = R(ind);
            S = S(ind);
            
            Ne = length(T);  % number of total splines in T,E,... arrays
            
            
            % Ensure R is nonzero:
            R = max(R, 1e-100); 

            %--------------------------------------------------------------
            % Find (T,R) point of max positive curvature
            %--------------------------------------------------------------
            Rpp = Compute_R_E_Curvature(R,T);

            [~,Ibest] = max(Rpp);  % index of point with max positive curvature

            Cold = C;  % (***) record old value
            C    = 5;  % number of choices of spline fit
            
            if  Ibest > Ne + 1 - C
                Ibest = Ne + 1 - C;  
            end

            %--------------------------------------------------------------
            % Generate C choices of possible spline fits
            %--------------------------------------------------------------
            fprintf('\n'),
            fprintf('Generating possible smoothing spline choices...\n'),

            DC   = true(1,C);             % Display Curves    flag
            DD   = true;                     % Display Data      flag

            Tsfc = T(Ibest:Ibest+C-1); % T of each spline fit choice
            Esfc = E(Ibest:Ibest+C-1); % E of each spline fit choice
            Rsfc = R(Ibest:Ibest+C-1); % R of each spline fit choice
            Ssfc = S(Ibest:Ibest+C-1); % S of each spline fit choice
            
            if Cold ~= C 
                yy = zeros(C,N);         % y(x)
                y1 = zeros(C,N);         %   dy/dx
                y2 = zeros(C,N);         % d^2y/dx^2
                y3 = zeros(C,N);         % d^3y/dx^3
            end
            
            for i = 1:C
                yy(i,:) = fnval(      Ssfc(i)   ,x);
                y1(i,:) = fnval(fnder(Ssfc(i),1),x);
                y2(i,:) = fnval(fnder(Ssfc(i),2),x);
                y3(i,:) = fnval(fnder(Ssfc(i),3),x);
            end

            %--------------------------------------------------------------
            % Record new values in "all values" arrays
            Tall(Nall+[1:Nnew]) = Tnew(:);   
            Rall(Nall+[1:Nnew]) = Rnew(:);   
            Eall(Nall+[1:Nnew]) = Enew(:);
            Nall = Nall + Nnew;
        
            %--------------------------------------------------------------
            % Delete existing plot data
            delete([HHyy,HHy1,HHy2,HHy3,HHro,HHer])

            % If C has changed, then need to reallocate memory
            if Cold ~= C
                HHyy = zeros(1,C);
                HHy1 = zeros(1,C);
                HHy2 = zeros(1,C);
                HHy3 = zeros(1,C);
                HHro = zeros(1,C);
                
                CLRlegend = CLRlegend_all(1:C); % Legend
            end             
            
            %--------------------------------------------------------------
            % Plot all new curves
            figure(Hro)
                set(HHmax,'visible','off');
                HHer = plot(T,R,'k.','MarkerSize',psize);
                for i = 1:C
                    HHro(i) = plot(Tsfc(i),Rsfc(i),'.','color',CLR(i,:),'MarkerSize',psize+2);
                end
                legend(HHro(DC),CLRlegend(DC),LegendStyle{:})
                set(gca,'XLim',[T(1) T(end)])
                
            figure(Hy)
                for i = 1:C
                    HHyy(i) = plot(x,yy(i,:),'color',CLR(i,:),'linewidth',1);    
                end
                legend([HHdata(DD),HHyy(DC)],[DATlabel(DD),CLRlegend(DC)],LegendStyle{:})
                
            figure(Hy1)
                for i = 1:C    
                    HHy1(i) = plot(x,y1(i,:),'color',CLR(i,:),'linewidth',1);
                end
            	legend(HHy1(DC),CLRlegend(DC),LegendStyle{:})
                
            figure(Hy2)
                for i = 1:C
                    HHy2(i) = plot(x,y2(i,:),'color',CLR(i,:),'linewidth',1);
                end
            	legend(HHy2(DC),CLRlegend(DC),LegendStyle{:})
                
            figure(Hy3)    
                for i = 1:C
                    HHy3(i) = plot(x,y3(i,:),'color',CLR(i,:),'linewidth',1);
                end
            	legend(HHy3(DC),CLRlegend(DC),LegendStyle{:})
            %--------------------------------------------------------------
        
        %------------------------------------------------------------------
        case {'n','N'}
            if length(action) == 1
                Ne = 20;
            else
                Ne = str2num(action(2:end));
            end

            %--------------------------------------------------------------
            % Select new T bounds
            disp(' ')
            disp('Mark lower and upper bounds for new tol range.')
            
            figure(Hro);
            
            [Tdes,Rdes] = ginput(2);

            % Ensure T are unique (and sorted in ascending order)
            Tdes = unique(Tdes);
            
            if length(Tdes)==1,  error('Must choose different tol values.'), end

            
            % Define range of error tolerance values
            T = logspace( log10(Tdes(1)), log10(Tdes(2)), Ne)';

            %==============================================================
            % CODE COPIED FROM INITIAL MANUAL MODE CODE BLOCK...except (***) items
            %==============================================================
            Cold = C;  % (***) record old value
            C    = 5;  % number of choices of spline fit
            
            %----------------------------------------------------------------------
            % Generate R-T frontier for smoothing splines 
            %----------------------------------------------------------------------
            fprintf('\n'),
            fprintf('Generating roughness-error frontier for smoothing splines ... \n'),

            %
            % (***)...EXCEPT the lines defining Ne and T were modified above.
            % 

            E = zeros(Ne,1);  % error
            R = zeros(Ne,1);  % roughness

            % define struct for B-spline data
            S       = struct('form','B-','knots',[],'coefs',[],'number',[],'order',[],'dim',1);
            S(Ne,1) = S(1);  % allocate memory for array of structs

            for i = Ne : -1 : 1
                fprintf('Evaluating spline %2d of %2d. \n',1+Ne-i,Ne),

                [R(i),S(i),E(i)] = FitSpline_ComputeRoughness(x,y,type,T(i));

                % Stop computing splines if R(i) is nearly Rmax
                if i < Ne+1-C && abs(Rmax - R(i)) < 0.1*Rmax
                    T = T(i:Ne);
                    E = E(i:Ne);
                    R = R(i:Ne);
                    S = S(i:Ne);
                    Ne = length(T);
                    fprintf('Roughness nearly maximum value.  Continuing...\n')
                    break,
                end
            end
    
            % Ensure R is nonzero:
            R = max(R, 1e-100); 

            %----------------------------------------------------------------------
            % Find (T,R) point of max positive curvature
            %----------------------------------------------------------------------
            Rpp = Compute_R_E_Curvature(R,T);

            [~,Ibest] = max(Rpp);  % index of point with max positive curvature

            if  Ibest > Ne + 1 - C
                Ibest = Ne + 1 - C;  
            end

            %----------------------------------------------------------------------
            % Generate C choices of possible spline fits
            %----------------------------------------------------------------------
            fprintf('\n'),
            fprintf('Generating possible smoothing spline choices...\n'),

            DC   = true(1,C);             % Display Curves    flag
            DD   = true;                     % Display Data      flag

            Tsfc = T(Ibest:Ibest+C-1); % T of each spline fit choice
            Esfc = E(Ibest:Ibest+C-1); % E of each spline fit choice
            Rsfc = R(Ibest:Ibest+C-1); % R of each spline fit choice
            Ssfc = S(Ibest:Ibest+C-1); % S of each spline fit choice
            
            if Cold ~= C  % (***) added if statement
                yy = zeros(C,N);         % y(x)
                y1 = zeros(C,N);         %   dy/dx
                y2 = zeros(C,N);         % d^2y/dx^2
                y3 = zeros(C,N);         % d^3y/dx^3
            end
            
            for i = 1:C
                yy(i,:) = fnval(      Ssfc(i)   ,x);
                y1(i,:) = fnval(fnder(Ssfc(i),1),x);
                y2(i,:) = fnval(fnder(Ssfc(i),2),x);
                y3(i,:) = fnval(fnder(Ssfc(i),3),x);
            end
    
            %==============================================================
            % END: CODE COPIED FROM INITIAL MANUAL MODE CODE BLOCK 
            %==============================================================
        

            %--------------------------------------------------------------
            % Record new values in "all values" arrays
            Tall(Nall+[1:Ne]) = T(:);   
            Rall(Nall+[1:Ne]) = R(:);   
            Eall(Nall+[1:Ne]) = E(:);
            Nall = Nall + Ne;
        
            %--------------------------------------------------------------
            % Delete existing plot data
            delete([HHyy,HHy1,HHy2,HHy3,HHro,HHer])

            % If C has changed, then need to reallocate memory
            if Cold ~= C
                HHyy = zeros(1,C);
                HHy1 = zeros(1,C);
                HHy2 = zeros(1,C);
                HHy3 = zeros(1,C);
                HHro = zeros(1,C);
                
                CLRlegend = CLRlegend_all(1:C); % Legend
            end             
            
            %--------------------------------------------------------------
            % Plot all new curves
            figure(Hro)
                set(HHmax,'visible','off');
                HHer = plot(T,R,'k.','MarkerSize',psize);
                for i = 1:C
                    HHro(i) = plot(Tsfc(i),Rsfc(i),'.','color',CLR(i,:),'MarkerSize',psize+2);
                end
                legend(HHro(DC),CLRlegend(DC),LegendStyle{:})
                set(gca,'XLim',[T(1) T(end)])
                
            figure(Hy)
                for i = 1:C
                    HHyy(i) = plot(x,yy(i,:),'color',CLR(i,:),'linewidth',1);    
                end
                legend([HHdata(DD),HHyy(DC)],[DATlabel(DD),CLRlegend(DC)],LegendStyle{:})
                
            figure(Hy1)
                for i = 1:C    
                    HHy1(i) = plot(x,y1(i,:),'color',CLR(i,:),'linewidth',1);
                end
            	legend(HHy1(DC),CLRlegend(DC),LegendStyle{:})
                
            figure(Hy2)
                for i = 1:C
                    HHy2(i) = plot(x,y2(i,:),'color',CLR(i,:),'linewidth',1);
                end
            	legend(HHy2(DC),CLRlegend(DC),LegendStyle{:})
                
            figure(Hy3)    
                for i = 1:C
                    HHy3(i) = plot(x,y3(i,:),'color',CLR(i,:),'linewidth',1);
                end
            	legend(HHy3(DC),CLRlegend(DC),LegendStyle{:})
            %--------------------------------------------------------------
        
        %------------------------------------------------------------------
        case {'d','D'}
            % Update whether original data is displayed
            figure(Hy),
            
            if (DD) 
                DD = false;
                set(HHdata  ,'visible','off')  
            else
                DD = true;
                set(HHdata  ,'visible','on')  
            end
            
            legend([HHdata(DD),HHyy(DC)],[DATlabel(DD),CLRlegend(DC)],LegendStyle{:})
            
            figure(Hro),          
               
        %------------------------------------------------------------------
        case {'q','Q'}
            sp  = NaN;
            R   = NaN;
            E   = NaN;
            tol = NaN;
            return
            
        %------------------------------------------------------------------
        otherwise
            disp('Command not recognized...try again.')
        end
    end

    %----------------------------------------------------------------------
    % Output selected smoothing spline
    %----------------------------------------------------------------------
    sp     = Ssfc(Cbest);
    sp.tol = Tsfc(Cbest);
    sp.E   = Esfc(Cbest);
    sp.R   = Rsfc(Cbest);


%===================================================== END MANUAL mode code   
end % switch mode

%--------------------------------------------------------------------------
% Discard unused array locations
Tall(Nall+1:end) = [];
Eall(Nall+1:end) = [];
Rall(Nall+1:end) = [];

% Overwrtite tol, E, and R with all values, now sorted into ascending order
[tol,ind] = unique(Tall);      % assign tol output array 
E         =        Eall(ind);
R         =        Rall(ind);


%--------------------------------------------------------------------------
% Output additional fields to smoothing spline data structure
%--------------------------------------------------------------------------
sp.x = x(:)';  % now a ROW vector
sp.y = y(:)';

% Compute s(x) and derivatives:
sp.s     = fnval(fnxtr(      sp   ), sp.x);
sp.d1sdx = fnval(fnxtr(fnder(sp,1)), sp.x);
sp.d2sdx = fnval(fnxtr(fnder(sp,2)), sp.x);
sp.d3sdx = fnval(fnxtr(fnder(sp,3)), sp.x);


end % function smoothspline
%==========================================================================

%==========================================================================
function [] = UpdatePlotVisibility(Hy,HHyy,HHdata, Hy1,HHy1, Hy2,HHy2, Hy3,HHy3, Hro, DC,DD,DATlabel,CLRlegend,LegendStyle)
    % Update which curves are displayed
    %----------------------------------------------------------------------
    figure(Hy),
        set(HHyy    ,'visible','off')
        set(HHyy(DC),'visible','on')
        legend([HHdata(DD),HHyy(DC)],[DATlabel(DD),CLRlegend(DC)],LegendStyle{:})

    figure(Hy1),
        set(HHy1    ,'visible','off')  
        set(HHy1(DC),'visible','on')    
        legend(HHy1(DC),CLRlegend(DC),LegendStyle{:})

    figure(Hy2),
        set(HHy2    ,'visible','off')  
        set(HHy2(DC),'visible','on')    
        legend(HHy2(DC),CLRlegend(DC),LegendStyle{:})

    figure(Hy3),
        set(HHy3    ,'visible','off')  
        set(HHy3(DC),'visible','on')    
        legend(HHy3(DC),CLRlegend(DC),LegendStyle{:})

    figure(Hro),
    %--------------------------------------------------------------
end % function UpdatePlotVisibility
%==========================================================================

%==========================================================================
function E = ComputeError(x,y,S)
    % Compute error of given spline
    E = trapz(x, ( y - fnval(S,x) ).^2 );   
    
    % % This error E is consistent with the weighted sum:
    %     m = S.order/2;
    %     [~,~,w] = Check_xy_Compute_w(x,y,m);
    %     Echeck = sum( w(:) .* ( y(:) - fnval(S,x(:)) ).^2 )
    %     E - Echeck
end
%==========================================================================

%==========================================================================
function R = ComputeRoughness(x,type,S)
    % Compute roughness of given spline
    R = trapz(x, fnval( fnder(S,type), x ).^2 );    
end
%==========================================================================

%==========================================================================
function [R,S,E] = FitSpline_ComputeRoughness(x,y,type,tol)
    % Fit smoothing spline with given error tolerance, tol
    S = spaps(x,y,tol,type);

    % Compute error and roughness
    E = ComputeError(x,y,S);
    R = ComputeRoughness(x,type,S);
end
%==========================================================================

%==========================================================================
function Rpp = Compute_R_E_Curvature(R,E)
    % Compute curvature of R-E curve.
    %----------------------------------------------------------------------
    % convert to log space:
    logE = log(E(:));
    logR = log(R(:));
    
    % %     %----------------------------------------------------------------------
    % %     % Original method:  Compute curvature, Rpp = d^2(log(R))/d(log(E))^2  
    % %     %----------------------------------------------------------------------
    % %     Rpp = 0*R(:);   % [n,1] column vector
    % %     for i = 2:length(E)-1
    % %        Rpp(i) = (  (logR(i+1)-logR(i  )) / (logE(i+1)-logE(i)  )   ...
    % %                  - (logR(i)  -logR(i-1)) / (logE(i)  -logE(i-1)) ) ... 
    % %                  / ...
    % %                  (0.5*(logE(i+1)-logE(i-1)));
    % %     end
    % %          
    % %     [~,Rpp_check] = FiniteDifferenceNonUniform(logE,logR);
    % %     
    % %     Rpp - Rpp_check
    
    
    %----------------------------------------------------------------------
    % 7/28/2021 BEPPS: Curvature of parameterized curve E(t),R(t)
    %----------------------------------------------------------------------
    % normalize to stencil width:
    logE = (logE - logE(1)  ) / (logE(end) - logE(1));
    logR = (logR - logR(end)) / (logR(1) - logR(end));
    
    % Compute arclength dt and curve parameter t
    dt = sqrt( diff(logE).^2 + diff(logR).^2 ); 
    t  = cumsum( [0; dt] );
    
    % Compute derivatives:
    %         d1E =   d(logE)/dt  
    %         d2E = d^2(logE)/dt^2
    % 
    %         d1R =   d(logR)/dt  
    %         d2R = d^2(logR)/dt^2
    %
    [d1E,d2E] = FiniteDifferenceNonUniform(t, logE);
    [d1R,d2R] = FiniteDifferenceNonUniform(t, logR);
    
    % Signed curvature: 
    Rpp = (d1E.*d2R - d1R.*d2E) ./ (d1E.^2 + d1R.^2).^(3/2);

    % %     fprintf('\n Curvature check:\n')
    % %     [~,Rpp_check] = FiniteDifferenceNonUniform(logE,logR);
    % %     
    % %     [Rpp, Rpp_check, Rpp - Rpp_check]
    % %     1;
    
end
%==========================================================================
    
%==========================================================================
function [dfdx,d2fdx2] = FiniteDifferenceNonUniform(x,f)
    %  Computes the first and second derivatives of given function f(x) 
    %  via the central difference method.  
    %
    %   f   [N,1]   dependent variable array
    %   x   [N,1] independent variable array (may have non-uniform spacing)
    %
    %   dfdx   = df/dx       first  derivative
    %   d2fdx2 = d^2f/dx^2   second derivative  **** NOTE:  d2fdx2(1)==d2fdx2(N)=0 
    %
    % Reference:  Hilding Sundqvist and George Veronis (1970) "A simple 
    % finite-difference grid with non-constant intervals", Tellus XXII.
    %
    %----------------------------------------------------------------------
    f = f(:);  % ensure vectors are shaped [N,1]
    x = x(:);
    
    N = length(x);  
    h = diff(x);  % step size, may be non-uniform    

    i = 2:N-1;

    dfdx = [ (- f(3).*  h(1).^2 ...
              + f(2).* (h(1)+h(2)).^2 ...
              - f(1).*[(h(1)+h(2)).^2 - h(1).^2] ) ./ ( h(1).*h(2).*(h(1)+h(2)) );   % forward difference for point i=1, O(h^2) error
             ... 
               f(i+1).*      h(i-1) ./(h(i)  .*(h(i-1)+h(i))) ...
             - f(i-1).* h(i)        ./(h(i-1).*(h(i-1)+h(i))) ...
             + f(i)  .*(h(i)-h(i-1))./(h(i)  .* h(i-1)      );    % central difference, O(h^2) error
             ...
             (+ f(N-2).*   h(N-1).^2 ...
              - f(N-1).*  (h(N-1)+h(N-2)).^2 ...
              + f(N)  .* [(h(N-1)+h(N-2)).^2 - h(N-1).^2] ) ./ ( h(N-1).*h(N-2).*(h(N-1)+h(N-2)) ) ];      % backwrd difference for point i=N, O(h^2) error

    d2fdx2 = [ 0; ...  % ignore endpoints
             ... 
               2*f(i+1) ./(h(i)  .*(h(i-1)+h(i))) ...
             + 2*f(i-1) ./(h(i-1).*(h(i-1)+h(i))) ...
             - 2*f(i)   ./(h(i)  .* h(i-1)      );    % central difference, O(h^2) error
             ...
             0 ];  % ignore endpoints
end
%==========================================================================

%==========================================================================
function [x,y,w] = Check_xy_Compute_w(x,y,m)
    %----------------------------------------------------------------------    
    % Perform error checking on the input x and y arrays (and modify them if necessary)
    %
    % Compute weights via trapezoidal rule, accounting for repeated x values.
    %----------------------------------------------------------------------
    
    %----------------------------------------------------------------------
    % Perform error checking on the input x and y arrays (and modify them if necessary)
    %--------------------------------------------------------------------------
    % ensure x and y are numeric arrays (not logical, character, string, cell, or structure arrays)
    if ~isnumeric(x) || ~isnumeric(y),  error('x and y must be numeric arrays.'),  end

    % ensure x and y are vectors (1-by-n or n-by-1):
    if  ~isvector(x) ||  ~isvector(y),  error('x and y must be vectors.'),     end

    % ensure x and y are the same size:
    if ~all( size(x) == size(y) ),     error('x and y must be the same size.'),  end

    % ensure x and y are real:
    if ~isreal(x) ||  ~isreal(y),  error('x and y must be real.'),  end


    % Remove NaN and Inf data (retain finite data)
    ind = isfinite(x) & isfinite(y);
    if ~all(ind)
        x  = x(ind);  
        y  = y(ind);
    end

    % Ensure x and y are column vectors
    if isrow(x)  % ...equivalent to... if size(x,1) == 1 && size(x,2) > 1
        x = x(:);
        y = y(:);
    end

    % Ensure x is nondecreasing
    if any( diff(x) < 0 )
        [x,ind] = sort(x); 
         y = y(ind);
    end

    % Record number of data points:
    N = length(x);  % x is size [N,1]
    Nrepeats = sum( diff(x)==0 );  % number of repeated x values

    % Ensure sufficient data points:
    if N - Nrepeats < m, error('insufficient data points'),  end



    %----------------------------------------------------------------------
    % Compute weights w consistent with Composite Trapezoidal Rule 
    %----------------------------------------------------------------------
    dx = diff(x);  % [N-1,1]  COLUMN vector
    
    w = 0.5*[dx(1); dx(1:end-1)+dx(2:end);  dx(end)]';   % [1,N] row vector
        
    %----------------------------------------------------------------------        
    % Combine repeated data sites: average y and sum w
    if Nrepeats > 0 
        % tolred = 0;  % "tolred" only needed for Matlab chkxywp implementation
        
        %------------------------------------------------------------------
        % Compute knot multiplicity as in:   mults2 = knt2mlt(x)
        % 
        % Assert: x is nondecreasing, so dx is nonnegative.
        % 
        mults = zeros(N,1);
        
        for j = 2:N
            if dx(j-1) == 0
                mults(j) = mults(j-1) + 1;
            end
        end
        %------------------------------------------------------------------
        
        %------------------------------------------------------------------
        % For each repeated x location...combine y and w.
        for j = find( diff([mults;0]') < 0 )

            j0  = j - mults(j);  % index of first repeated x point
            ind = j0:j;          % indices of repeated x points
            
            % %--------------------------------------------------------------
            % %--------------------------------------------------------------
            % % This is the original Matlab chkxywp.m implementation.  It has
            % % a bug: When x is repeated more than once, then w is zero for 
            % % the "interior" x values. e.g.
            % %
            % %            x         w     mults
            % %         0.0000    0.6500    0
            % %         1.3000    1.0000    0
            % %         2.0000    0.8500    0
            % %         3.0000    0.5000    0
            % %         3.0000         0    1   <-- "interior" point ignored
            % %         3.0000         0    2   <-- "interior" point ignored
            % %         3.0000    0.6000    3
            % %         4.2000    1.1000    0
            % %         5.2000    0.5000    0
            % %
            % % The matrix multiplications w(ind)*y(ind) do not include the
            % % "interior" y values.  Thus, the "y(j0)" value and "tolred"
            % % variable are junk.
            % temp = sum( w(ind) ); % sum weights
            % 
            %                 % row * column = scalar
            % tolred = tolred + w(ind)*y(ind).^2;  
            % 
            % y(j0) = (w(ind)*y(ind))/temp;
            % w(j0) = temp;
            % 
            % tolred = tolred - temp * y(j0)^2;
            % 
            % % Since this "tolred" value is junk anyway, I choose to simply
            % % ignore it and define the error for this point as that between
            % % the spline and the average y value.
            % %--------------------------------------------------------------
            % %--------------------------------------------------------------
            
            %--------------------------------------------------------------
            % BEPPS implementation: Use actual mean y value, and ignore tolred
            y(j0) = mean( y(ind) );
            w(j0) =  sum( w(ind) ); 
            %--------------------------------------------------------------
        end

        ind_repeats = find(mults ~= 0);
        
        x(ind_repeats) = []; 
        y(ind_repeats) = []; 
        w(ind_repeats) = []; 

    end    
end % function Check_xy_Compute_w
%==========================================================================

 