%-------------------------------------------------------------------------- 
% smoothspline_example.m
%
% Script to run example case for smoothspline.m
% 
%-------------------------------------------------------------------------- 
% Example Setup: 
% 
% Define a function to analyze
x  = 0:0.01:10;   
yt = exp(-x).*sin(x);                     % true  y(x)
y  = yt +  0.001*randn(size(x));          % noisy y(x)

type  = 'quintic';  % choose 'cubic' or 'quintic'


% ---------------------
% Example: Auto mode
 
[sp,R,E,tol] = smoothspline(x,y,type,'auto',true);


% ---------------------
%% Example: Manual mode
 
[sp,R,E,tol] = smoothspline(x,y,type,'manual');

%   A roughness-error frontier is computed and plotted, and the user is
%   prompted to enter one of the following text commands:
%
%           h ii   hide spline(s)
%           v ii   view spline(s)
%           p      pick new splines to view
%           n (i)  new... error tol range      (default: 20)
%           r (i)  refine error tol resolution (default: 1)
%           d      data plot toggle
%           s i    select spline
%           q      quit
%
% 
%   By way of example,
%
%   >> h125     hides splines 1, 2, and 5.
%
%   >> v12      displays splines 1 and 2.  Vii is the opposite of Hii, and
%               ii can be any one or more of the current spline numbers.
% 
%   >> p        pick new splines: user is prompted to click on points in
%               the Roughness-Error figure, and those splines are shown. 
%  
%   >> n        new error tolerance range:  the user selects two error
%               tolerance values (from the R-E figure), and 20 new splines 
%               in this range are computed.  To specify the number of new
%               splines, say 30, use the command: n30.
%  
%   >> r        refine error tolerance resolution:  the user selects two  
%               error tolerance values (from the R-E figure), and 1 new  
%               spline is inserted between each of the splines in this range.   
%               To specify the number of new splines to insert, say 2, use 
%               the command: r2.
%  
%   >> d        toggles whether the original x,y data is shown.
%
%   >> s3       selects spline 3 for output. This terminates the program.
% 
%   >> q        quit: terminates the program with all outputs given as NaN.


% ---------------------
%% Example: Postprocessing the output results

% Check error and roughness of output spline
Echeck = trapz(x,       ( y - fnval(sp,x) ).^2 )       % error     of sp
Rcheck = trapz(x, fnval(fnder(sp,sp.order/2),x).^2 )   % roughness of sp
Echeck - sp.E
Rcheck - sp.R

% Compare output with analytic derivatives and roughness (for the above example):
d1ytdx =   -exp(-x).*sin(x) +   exp(-x).*cos(x);    %   dyt/dx
d2ytdx = -2*exp(-x).*cos(x);                        % d^2yt/dx^2
d3ytdx =  2*exp(-x).*cos(x) + 2*exp(-x).*sin(x);    % d^3yt/dx^3
Rt2 = trapz(x, d2ytdx.^2)  % compare to cubic   spline roughness
Rt3 = trapz(x, d3ytdx.^2)  % compare to quintic spline roughness 

% Plot results
figure(5), clf, box on, grid on, loglog(tol,R,'k.','MarkerSize',18)  
               hold on,   loglog(sp.tol, sp.R,'r.','MarkerSize',22)

figure(4), clf, box on, grid on, plot(x,y,'.b',  x,yt,'k-', sp.x,sp.s,'r--')
figure(1), clf, box on, grid on, plot(x,d1ytdx,'k-',  sp.x,sp.d1sdx,'r--')
figure(2), clf, box on, grid on, plot(x,d2ytdx,'k-',  sp.x,sp.d2sdx,'r-.')
figure(3), clf, box on, grid on, plot(x,d3ytdx,'k-',  sp.x,sp.d3sdx,'r-.')